; SketchyLISP -- An interpreter for purely applicative Scheme
; Copyright (C) 2006,2007 Nils M Holm <nmh@t3x.org>
; See the file LICENSE for conditions of use.

(package 'core)

; Syntax transformer

; (define (d x) (begin (display "D--> ") (write x) (newline) x))

; Extend environment
(define (ext-env name x v env)
  (cond ((assq x env)
      (bottom (append (list name ':)
                      '(duplicate variable in pattern:)
                      (list x))))
    (else (cons (cons x v) env))))

; Match an ellipsis.
; HEAD is the part to be associated with the ellipsis.
; TAIL is the remaining form to match after the ellipsis.
; Thus function uses a longest match first approach, so
; (match-ellipsis '(x x k k k) '(k) 'n '(k) ())
;   => ((... . (x x k k)))

(define (match-ellipsis form pattern name lits env)
  (letrec
    ((try-match (lambda (head tail)
      (let ((v (match tail pattern name lits env)))
        (cond (v (ext-env name '... (reverse head) v))
          ((null? head) #f)
          (else (try-match (cdr head)
                (cons (car head) tail))))))))
    (try-match (reverse form) ())))

; Match form against pattern.
; name - name of syntax transformer;
; lits - literals (keywords) of syntax-rules;
; env - initial environment.
; Returns an environment with bindings of
; pattern variables to subforms or #F in
; cass form does not match pattern.

(define (match form pattern name lits env)
  (letrec
    ((_match (lambda (form pattern env)
      (cond ((eq? pattern '_)
          (if (eq? form name) env #f))
        ((memq pattern lits)
          (if (eq? form pattern) env #f))
        ((and (pair? pattern) (eq? (car pattern) '...))
          (match-ellipsis form (cdr pattern)
                          name lits env))
        ((symbol? pattern)
          (if (memq pattern lits)
              (if (eq? pattern form) env #f)
              (ext-env name pattern form env)))
        ((and (pair? pattern)
              (pair? form))
          (let ((e (_match (car form) (car pattern) env)))
            (and e (_match (cdr form) (cdr pattern) e))))
        (else (if (equal? form pattern) env #f))))))
    (_match form pattern env)))

; Find rule with a pattern that matches the given form.
; rules - list of rules to try;
; name - name of syntax transformer;
; lits - literals (keywords) of syntax-rules.

(define (find-rule form rules name lits)
  (cond ((null? rules)
      (bottom name ': 'bad 'syntax))
    (else (let ((e (match form (caar rules) name lits '())))
          (cond (e (list (caar rules) (cadar rules) e))
            (else (find-rule form (cdr rules) name lits)))))))

; Replace OLD by NEW in FORM.
(define (replace old new form)
  (cond ((equal? form old) new)
    ((pair? form)
      (cons (replace old new (car form))
            (replace old new (cdr form))))
    (else form)))

; Substitute keys of A by values of A in X.
(define (substitute x a)
  (let ((v (assoc x a)))
    (cond (v (cdr v))
      ((pair? x)
        (cons (substitute (car x) a)
              (substitute (cdr x) a)))
      (else x))))

; Substitute forms matching an ellipsis
(define (subst-rest rest ptrn form)
  (cond ((pair? ptrn)
      (let ((p (car ptrn))
            (f (car form)))
        (map (lambda (x)
               (let ((e (cons x (replace p x f))))
                 (substitute x (list e))))
             rest)))
    (else rest)))

; Substitute variables of env by values of env in form.
(define (tmpl->form ptrn form env)
  (cond ((not (pair? form))
      (let ((v (assv form env)))
        (if v (cdr v) form)))
    ((and (pair? form) (pair? (cdr form)) (eq? (cadr form) '...))
      (let ((new (tmpl->form #f (car form) env))
            (v (assq '... env)))
        (if v (append (list new)
                      (subst-rest (cdr v) ptrn form)
                      (cddr form))
              (bottom 'unmatched '... 'in 'syntax-rule))))
    ((and (pair? form) (eq? (car form) '...))
      (bottom '(syntax-rules: this should not happen)))
    ((pair? form)
      (cons (tmpl->form (if (pair? ptrn) (car ptrn) #f) (car form) env)
            (tmpl->form (if (pair? ptrn) (cdr ptrn) #f) (cdr form) env)))
    (else form)))

; Transform the syntax of the given form.
; This is the entry point of the syntax transformer
; when called from the SketchyLISP core.

(define (transform-syntax form)
  (let ((syn (syntax->list (car form))))
    (if (not syn)
        (bottom 'bad 'syntax 'rules: form)
        (let ((name (car syn))
              (lits (cadr syn))
              (rules (caddr syn)))
          (let ((pat/rule/env (find-rule form rules name lits)))
            (transform-all
              (apply tmpl->form pat/rule/env)))))))

; TRANSFORM-ALL *must* be defined *after* TRANSFORM-SYNTAX.
; This is probably because TRANSFORM-SYNTAX is defined by
; the SketchyLISP core. To be investigated!

; Syntax-transform all subforms of the given form.
(define (transform-all form)
  (cond ((not (pair? form)) form)
    ((eq? (car form) 'quote) form)
    ((syntax->list (car form))
      (transform-syntax form))
    (else (map transform-all form))))

(package)
